#!/bin/bash
#
# A service script to dump the relevant MongoDB database(s)
# of a docker-composed MongoDB instance. Creates a tarball in
# $BASE_DIR/storage/backups/tarballs folder (by default).
# The relevant databases must be specified within the
# docker-compose.yml in a BACKUP_DATABASES environment variable.
# An optional parameter may change the target folder.
#
# Call as a Docker manager user (member of the docker Linux group) via cron.
#
# Author: Kovács Zoltán <kovacsz@marcusconsulting.hu>
# License: GNU/GPL 3+ https://www.gnu.org/licenses/gpl-3.0.en.html
# 2025-03-05 v0.1.1
# mod: minimally rewrited the description.
# 2024-09-23 v0.1 Initial version.

# Accepted environment variables and their defaults.
#
PAR_BASEDIR=${PAR_BASEDIR:-""}                  # Service's base folder
PAR_BACKUPDIR=${PAR_BACKUPDIR:-""}              # Folder to dump within

# Messages (maybe overridden by configuration).
#
MSG_DOCKERGRPNEED="You must be a member of the docker group."
MSG_DOESNOTRUN="This service doesn't run."
MSG_MISSINGDEP="Fatal: missing dependency"
MSG_MISSINGYML="Fatal: didn't find the docker-compose.yml file"
MSG_NONWRITE="The target directory isn't writable"
MSG_NOLOCATE="Cannot locate the MongoDB container."
MSG_NOPARAM="Missing environment parameter"


# Other initialisations.
#
BACKUPDIR="storage/backups/tarballs"		# Folder to dump within
PAR_DATABASES="DS_BACKUP_DATABASES"		# List of DB(s) in YMLFILE
SERVICENAME="mongodb"				# The composed MongoDB service
USER=${USER:-LOGNAME}                           # Fix for cron enviroment only
YMLFILE="docker-compose.yml"			# Gets the parameters from here

# Checks the dependencies.
#
TR=$(which tr 2>/dev/null)
if [ -z "$TR" ]; then echo "$MSG_MISSINGDEP tr."; exit 1 ; fi
for item in basename cat cut date dirname docker \
            grep gzip hostname id pwd tail xargs
do
    if [ -n "$(which $item)" ]
    then export $(echo $item | "$TR" '[:lower:]' '[:upper:]' | "$TR" '-' '_')=$(which $item)
    else echo "$MSG_MISSINGDEP $item." >&2; exit 1; fi
done
# All dependencies are available via "$THECOMMAND" (upper case) call.
#
# Let's find which version of docker-compose is installed.
if [ $($DOCKER compose version 2>&1 >/dev/null; echo $?) -eq 0 ]; then
    # We'll use v2 if it is available.
    DOCKER_COMPOSE="$DOCKER"
    commandstring="compose"
else
    # Otherwise falling back to v1.
    DOCKER_COMPOSE="$(which docker-compose)"
    commandstring=""
fi
# One of the two is mandatory.
if [ -z "$DOCKER_COMPOSE" ];then echo "$MSG_MISSINGDEP docker-compose" >&2; exit 1; fi
# Below docker-compose should be called as "$DOCKER_COMPOSE" $commandstring sequence.
#
# An additional bugfix (use "$(which gzip)" instead of "$GZIP"):
# https://www.gnu.org/software/gzip/manual/html_node/Environment.html
GZIP=""

# Where I'm?
# https://gist.github.com/TheMengzor/968e5ea87e99d9c41782
SOURCE="$0"
while [ -h "$SOURCE" ]; do
    # resolve $SOURCE until the file is no longer a symlink
    SCRPATH="$( cd -P "$("$DIRNAME" "$SOURCE" )" && echo "$PWD" )" #"
    SOURCE="$("$READLINK" "$SOURCE")"
    # if $SOURCE was a relative symlink, we need to resolve it
    # relative to the path where the symlink file was located
    [[ $SOURCE != /* ]] && SOURCE="$SCRPATH/$SOURCE"
done; SCRPATH="$( cd -P "$("$DIRNAME" "$SOURCE" )" && echo "$PWD" )" #"

# Need to be root or a Docker manager user.
#
[[ "$USER" != 'root' ]] \
&& [[ -z "$(echo "$("$ID" -Gn "$USER") " | "$GREP" ' docker ')" ]] \
&& echo "$MSG_DOCKERGRPNEED" >&2 && exit 1 #"

# Searches the base folder, containing a docker-compose.yml file.
#
# Called from the base folder (./)?
BASE_DIR="$PAR_BASEDIR"
TEST_DIR="$SCRPATH"
[[ -z "$BASE_DIR" ]] && [[ -r "$TEST_DIR/$YMLFILE" ]] && BASE_DIR="$TEST_DIR"
# Called from ./tools?
TEST_DIR="$("$DIRNAME" "$TEST_DIR")"
[[ -z "$BASE_DIR" ]] && [[ -r "$TEST_DIR/$YMLFILE" ]] && BASE_DIR="$TEST_DIR"
# Called from ./tools/*.d?
TEST_DIR="$("$DIRNAME" "$TEST_DIR")"
[[ -z "$BASE_DIR" ]] && [[ -r "$TEST_DIR/$YMLFILE" ]] && BASE_DIR="$TEST_DIR"
# On failure gives it up here.
if [ -z "$BASE_DIR" -o ! -r "$BASE_DIR/$YMLFILE" ]; then
    echo "$MSG_MISSINGYML" >&2; exit 1
fi
# Sets the absolute paths.
BACKUPDIR="${PAR_BACKUPDIR:-$BASE_DIR/$BACKUPDIR}"

# The dump target folder must be writable.
#
[[ ! -w "$BACKUPDIR" ]] \
&& echo "$MSG_NONWRITE: $BACKUPDIR" >&2 && exit 1

# The service must be running - silently gives up here if not.
#
[[ -z "$(cd "$BASE_DIR"; "$DOCKER_COMPOSE" $commandstring ps --services --filter "status=running")" ]] \
&& exit 1

# Converts the MongoDB service name to an actual running container's name.
#
MDBCONTAINER="$("$DOCKER" inspect -f '{{.Name}}' $(cd "$BASE_DIR"; "$DOCKER_COMPOSE" $commandstring ps -q "$SERVICENAME") | "$CUT" -c2-)"
# Gives up here if failed.
if [ -z "$MDBCONTAINER" ]; then echo "$MSG_NOLOCATE" >&2; exit 1; fi

# Checks and parses the config file for database names to dump
# and DBA (root) credentials for MongoDB.
#
function parse { [[ -z "$1" ]] && return
    # Gets the live lines containing the parameter.
    value=$("$CAT" "$CONFFILE" 2>/dev/null | "$GREP" -ve '^#' | \
            "$GREP" -e " $1:" | "$TR" -d '\r')
    # If multiple the last one to consider.
    value=$(echo -e  "$value" | "$TAIL" -n1)
    # Right side of the equal sign W/O leading and trailing spaces and quotes.
    value=$(echo -ne "$value" | "$CUT" -d':' -f2 | "$XARGS")
    echo -e "$value"; return
}
# Examines the YMLFILE.
CONFFILE="$BASE_DIR/$YMLFILE"
# Gets the list of the databases to dump. Silently exits if it is empty.
DATABASES="$(parse "$PAR_DATABASES")"
if [ -z "$DATABASES" ]; then exit; fi
# All parameters below are mandatories.
DBAUTH="$(parse "MONGO_INITDB_DATABASE")"
if [ -z "$DBAUTH" ]; then echo "$MSG_NOPARAM MONGO_INITDB_DATABASE" >&2; exit 1; fi
DBUSER="$(parse "MONGO_INITDB_ROOT_USERNAME")"
if [ -z "$DBAUTH" ]; then echo "$MSG_NOPARAM MONGO_INITDB_ROOT_USERNAME" >&2; exit 1; fi
DBPASS="$(parse "MONGO_INITDB_ROOT_PASSWORD")"
if [ -z "$DBAUTH" ]; then echo "$MSG_NOPARAM MONGO_INITDB_ROOT_PASSWORD" >&2; exit 1; fi
# We've the configuration parsed.

# Attempts the dump(s) using the mongodump utility existing within the container.
# Uses the DBA (root) credentials parsed from the YMLFILE above.
#
if [ -w "$BACKUPDIR" ]; then
    # Enumerates the relevant databases (if any).
    for DBNAME in $DATABASES ''
    do
	# Dumps the actual database as a DBA.
	if [ -n "$DBNAME" ]; then
	    BACKUP_NAME=$SERVICENAME-$DBNAME.$("$DATE" '+%Y%m%d_%H%M%S').$("$HOSTNAME")
	    "$DOCKER" exec $MDBCONTAINER sh -c "exec mongodump -u $DBUSER -p $DBPASS --authenticationDatabase $DBAUTH -d $DBNAME --quiet --archive" | \
	    "$(which gzip)" > "$BACKUPDIR/$BACKUP_NAME.archive.gz" 2>>"$BACKUPDIR/$BACKUP_NAME.log"
	fi
    done
fi

# That's all, Folks! :)
