#!/bin/bash
#
# Launches a dockerized service defined by a docker-compose.yml file.
# This is a general purpose launcher, doesn't requires customization.
# Actually it is a humble wrapper script for a 'docker-compose up -d'
# command. Additionally may set up an aggregated logfile and rejects
# all multiple launch attempts.
#
# Author: Kovács Zoltán <kovacsz@marcusconsulting.hu>
#         Kovács Zoltán <kovacs.zoltan@smartfront.hu>
# License: GNU/GPL v3+ (https://www.gnu.org/licenses/gpl-3.0.en.html)
#
# 2025-02-02 v1.3
# mod: Only handles the aggregate log, per-service logging has been
#      outsourced to a new code snippet.
# 2025-02-02 v1.2
# fix: Unwanted error message in old docker-compose detection.
# 2024-08-24 v1.1
# new: docker-compose v2 compatibility - tested with Ubuntu 24.04 LTS.
# 2023-06-18 v1.0
# new: forked from the "Smartfront's DOCKER_skeleton" repository.
# 2021-09-02 v0.1 Initial release

# Accepted environment variables and their defaults.
PAR_BASEDIR=${PAR_BASEDIR:-""}			# Service's base folder
PAR_AGGLOGS=${PAR_AGGLOGS:-""}			# Not empty for aggregated logs

# Other initialisations.
LOGDIR="logs"
LOGFILE="logs/docker-compose.log"		# For aggregated log
YMLFILE="docker-compose.yml"

# Messages.
MSG_ALREADYRUN="This service is already running."
MSG_MISSINGDEP="Fatal: missing dependency"
MSG_MISSINGYML="Fatal: didn't find the docker-compose.yml file"

# Checks the dependencies.
TR=$(which tr 2>/dev/null)
if [ -z "$TR" ]; then echo "$MSG_MISSINGDEP tr."; exit 1 ; fi
for item in dirname docker readlink
do
    if [ -n "$(which $item)" ]
    then export $(echo $item | "$TR" '[:lower:]' '[:upper:]' | "$TR" '-' '_')=$(which $item)
    else echo "$MSG_MISSINGDEP $item." >&2; exit 1; fi
done
# All dependencies are available via "$THECOMMAND" (upper case) call.
#
# Let's find which version of docker-compose is installed.
if [ $($DOCKER compose version >/dev/null 2>&1; echo $?) -eq 0 ]; then
    # We'll use v2 if it is available.
    DOCKER_COMPOSE="$DOCKER"
    commandstring="compose"
else
    # Otherwise falling back to v1.
    DOCKER_COMPOSE="$(which docker-compose)"
    commandstring=""
fi
# One of the two is mandatory.
if [ -z "$DOCKER_COMPOSE" ];then echo "$MSG_MISSINGDEP docker-compose" >&2; exit 1; fi
# Below docker-compose should be called as "$DOCKER_COMPOSE" $commandstring sequence.

# Where I'm?
# https://gist.github.com/TheMengzor/968e5ea87e99d9c41782
SOURCE="$0"
while [ -h "$SOURCE" ]; do
    # resolve $SOURCE until the file is no longer a symlink
    SCRPATH="$( cd -P "$("$DIRNAME" "$SOURCE" )" && pwd )" #"
    SOURCE="$("$READLINK" "$SOURCE")"
    # if $SOURCE was a relative symlink, we need to resolve it
    # relative to the path where the symlink file was located
    [[ $SOURCE != /* ]] && SOURCE="$SCRPATH/$SOURCE"
done; SCRPATH="$( cd -P "$("$DIRNAME" "$SOURCE" )" && pwd )" #"

# Searches the base folder, containing a docker-compose.yml file.
# Called from the base folder (./)?
BASE_DIR="$PAR_BASEDIR"
TEST_DIR="$SCRPATH"
[[ -z "$BASE_DIR" ]] && [[ -r "$TEST_DIR/$YMLFILE" ]] && BASE_DIR="$TEST_DIR"
# Called from ./tools?
TEST_DIR="$("$DIRNAME" "$TEST_DIR")"
[[ -z "$BASE_DIR" ]] && [[ -r "$TEST_DIR/$YMLFILE" ]] && BASE_DIR="$TEST_DIR"
# Called from ./tools/*.d?
TEST_DIR="$("$DIRNAME" "$TEST_DIR")"
[[ -z "$BASE_DIR" ]] && [[ -r "$TEST_DIR/$YMLFILE" ]] && BASE_DIR="$TEST_DIR"
# On failure gives it up here.
if [ -z "$BASE_DIR" -o ! -r "$BASE_DIR/$YMLFILE" ]; then
    echo "$MSG_MISSINGYML" >&2; exit 1
fi

# Doesn't start if it is already running.
if [ -n "$(cd "$BASE_DIR"; "$DOCKER_COMPOSE" $commandstring ps --services --filter "status=running")" ]; then
    echo "$MSG_ALREADYRUN" >&2; exit 1
fi
# Starts the service.
(cd "$BASE_DIR"; "$DOCKER_COMPOSE" $commandstring up -d)

# Starts the aggregated logger on request - this process will automatically terminate
# when the composition stops.
if [ -n "$PAR_AGGLOGS" ]; then
    # Aggregated logs
    (cd "$BASE_DIR"; "$DOCKER_COMPOSE" $commandstring logs --no-color -t -f >> "$BASE_DIR/$LOGFILE" &)
fi

# That's all, Folks!
